import 'dart:async';
import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:shared_preferences/shared_preferences.dart';

import 'package:sudoku/entities/cell.dart';
import 'package:sudoku/utils/tools.dart';

typedef Board = List<List<Cell>>;
typedef ConflictsCount = List<List<int>>;
typedef AnimatedBoard = List<List<bool>>;
typedef AnimatedBoardSequence = List<AnimatedBoard>;

class Data extends ChangeNotifier {
  // Configuration available values
  final List<String> _availableParameters = ['level', 'size', 'skin'];

  final List<String> _availableLevelValues = ['easy', 'medium', 'hard', 'nightmare'];
  final List<String> _availableSizeValues = ['2x2', '3x2', '3x3', '4x4'];
  final List<String> _availableSkinValues = ['default', 'food', 'nature', 'monsters'];
  final List<String> _shufflableSkins = ['food', 'nature', 'monsters'];

  List<String> get availableParameters => _availableParameters;
  List<String> get availableLevelValues => _availableLevelValues;
  List<String> get availableSizeValues => _availableSizeValues;
  List<String> get availableSkinValues => _availableSkinValues;

  // Application default configuration
  String _parameterLevel = '';
  final String _parameterLevelDefault = 'medium';
  String _parameterSize = '';
  final String _parameterSizeDefault = '3x3';
  String _parameterSkin = '';
  final String _parameterSkinDefault = 'default';

  // Application current configuration
  String get parameterLevel => _parameterLevel;
  String get parameterSize => _parameterSize;
  String get parameterSkin => _parameterSkin;

  // Game data
  bool _assetsPreloaded = false;
  bool _gameIsRunning = false;
  bool _animationInProgress = false;
  int _blockSizeVertical = 0;
  int _blockSizeHorizontal = 0;
  Board _board = [];
  Board _boardSolved = [];
  ConflictsCount _boardConflicts = [];
  AnimatedBoard _boardAnimated = [];
  List<int> _shuffledCellValues = [];
  int? _selectedCellCol;
  int? _selectedCellRow;
  int? _selectedCellValue;
  bool _showConflicts = false;
  int _givenTipsCount = 0;
  int _givenTipsCountEnableCountdown = 0;
  String _currentSavedState = '';

  void updateParameterLevel(String parameterLevel) {
    _parameterLevel = parameterLevel;
    notifyListeners();
  }

  int get blockSizeVertical => _blockSizeVertical;
  int get blockSizeHorizontal => _blockSizeHorizontal;
  void updateParameterSize(String parameterSize) {
    _parameterSize = parameterSize;
    _blockSizeHorizontal = int.parse(_parameterSize.split('x')[0]);
    _blockSizeVertical = int.parse(_parameterSize.split('x')[1]);
    notifyListeners();
  }

  void updateParameterSkin(String parameterSkin) {
    _parameterSkin = parameterSkin;
    notifyListeners();
  }

  String getParameterValue(String parameterCode) {
    switch (parameterCode) {
      case 'level':
        return _parameterLevel;
      case 'size':
        return _parameterSize;
      case 'skin':
        return _parameterSkin;
    }
    return '';
  }

  List<String> getParameterAvailableValues(String parameterCode) {
    switch (parameterCode) {
      case 'level':
        return _availableLevelValues;
      case 'size':
        return _availableSizeValues;
      case 'skin':
        return _availableSkinValues;
    }
    return [];
  }

  void setParameterValue(String parameterCode, String parameterValue) async {
    switch (parameterCode) {
      case 'level':
        updateParameterLevel(parameterValue);
        break;
      case 'size':
        updateParameterSize(parameterValue);
        break;
      case 'skin':
        updateParameterSkin(parameterValue);
        break;
    }
    final prefs = await SharedPreferences.getInstance();
    prefs.setString(parameterCode, parameterValue);
  }

  void initParametersValues() async {
    final prefs = await SharedPreferences.getInstance();
    setParameterValue('level', prefs.getString('level') ?? _parameterLevelDefault);
    setParameterValue('size', prefs.getString('size') ?? _parameterSizeDefault);
    setParameterValue('skin', prefs.getString('skin') ?? _parameterSkinDefault);
  }

  String get currentSavedState => _currentSavedState;

  String computeCurrentGameState() {
    String cellsValues = '';
    const String stringValues = '0123456789ABCDEFG';
    for (int rowIndex = 0; rowIndex < _board.length; rowIndex++) {
      for (int colIndex = 0; colIndex < _board[rowIndex].length; colIndex++) {
        cellsValues += stringValues[_board[rowIndex][colIndex].value];
        cellsValues += _board[rowIndex][colIndex].isFixed ? 'x' : ' ';
      }
    }

    return json.encode({
      'level': _parameterLevel,
      'size': _parameterSize,
      'skin': _parameterSkin,
      'tipsCount': _givenTipsCount,
      'showConflicts': _showConflicts,
      'boardValues': cellsValues,
      'shuffledCellValues': _shuffledCellValues,
    });
  }

  void saveCurrentGameState() async {
    if (_gameIsRunning) {
      _currentSavedState = computeCurrentGameState();

      final prefs = await SharedPreferences.getInstance();
      prefs.setString('savedState', _currentSavedState);
    } else {
      resetCurrentSavedState();
    }
  }

  void resetCurrentSavedState() async {
    _currentSavedState = '';

    final prefs = await SharedPreferences.getInstance();
    prefs.setString('savedState', _currentSavedState);
    notifyListeners();
  }

  void loadCurrentSavedState() async {
    final prefs = await SharedPreferences.getInstance();
    _currentSavedState = prefs.getString('savedState') ?? '';
  }

  bool hasCurrentSavedState() {
    return (_currentSavedState != '');
  }

  Map<String, dynamic> getCurrentSavedState() {
    if (_currentSavedState != '') {
      Map<String, dynamic> savedState = json.decode(_currentSavedState);
      if (savedState.isNotEmpty) {
        return savedState;
      }
    }
    return {};
  }

  bool get gameIsRunning => _gameIsRunning;
  void updateGameIsRunning(bool gameIsRunning) {
    _gameIsRunning = gameIsRunning;
    notifyListeners();
  }

  bool get assetsPreloaded => _assetsPreloaded;
  void updateAssetsPreloaded(bool assetsPreloaded) {
    _assetsPreloaded = assetsPreloaded;
  }

  Board get board => _board;
  void updateCells(Board board) {
    _board = board;
    notifyListeners();
  }

  Board get boardSolved => _boardSolved;
  void updateCellsSolved(Board boardSolved) {
    _boardSolved = boardSolved;
  }

  AnimatedBoard get boardAnimated => _boardAnimated;
  ConflictsCount get boardConflicts => _boardConflicts;

  void shuffleCellValues() {
    const int maxCellValue = 16;
    final List<int> values = List<int>.generate(maxCellValue, (i) => i + 1);

    if (_shufflableSkins.contains(_parameterSkin)) {
      values.shuffle();
      printlog('Shuffled tiles values: $values');
    }

    _shuffledCellValues = values;
  }

  void setShuffleCellValues(List<int> values) {
    _shuffledCellValues = values;
  }

  int getTranslatedValueForDisplay(int originalValue) {
    return _shuffledCellValues[originalValue - 1];
  }

  int? get selectedCellCol => _selectedCellCol;
  set updateCelectedCellCol(int? selectedCellCol) {
    _selectedCellCol = selectedCellCol;
    notifyListeners();
  }

  int? get selectedCellRow => _selectedCellRow;
  set updateSelectedCellRow(int? selectedCellRow) {
    _selectedCellRow = selectedCellRow;
    notifyListeners();
  }

  int? get selectedCellValue => _selectedCellValue;
  set updateSelectedCellValue(int? selectedCellValue) {
    _selectedCellValue = selectedCellValue;
    notifyListeners();
  }

  int get givenTipsCountEnableCountdown => _givenTipsCountEnableCountdown;

  int get givenTipsCount => _givenTipsCount;
  void increaseGivenTipsCount() {
    _givenTipsCount = _givenTipsCount + 1;
    _givenTipsCountEnableCountdown = 60;
    const Duration interval = Duration(milliseconds: 500);
    Timer.periodic(
      interval,
      (Timer timer) {
        if (_givenTipsCountEnableCountdown == 0) {
          timer.cancel();
          notifyListeners();
        } else {
          _givenTipsCountEnableCountdown--;
          notifyListeners();
        }
      },
    );

    saveCurrentGameState();
    notifyListeners();
  }

  void setGivenTipsCount(int value) {
    _givenTipsCount = value;
    _givenTipsCountEnableCountdown = 0;
    notifyListeners();
  }

  bool canGiveTip() {
    return (_givenTipsCountEnableCountdown == 0);
  }

  void resetGivenTipsCount() {
    setGivenTipsCount(0);
  }

  void selectCell(int? col, int? row) {
    _selectedCellCol = col;
    _selectedCellRow = row;
    _selectedCellValue = null;
    if (row != null && col != null) {
      if (_board[row][col].value != 0) {
        _selectedCellValue = _board[row][col].value;
      }
    }
    notifyListeners();
  }

  void initConflictsBoard() {
    ConflictsCount nonConflictedBoard = [];
    final int boardSideLength = _blockSizeHorizontal * _blockSizeVertical;
    for (int row = 0; row < boardSideLength; row++) {
      List<int> line = [];
      for (int col = 0; col < boardSideLength; col++) {
        line.add(0);
      }
      nonConflictedBoard.add(line);
    }
    _boardConflicts = nonConflictedBoard;
  }

  void updateConflicts(ConflictsCount conflictsCount) {
    _boardConflicts = conflictsCount;
  }

  void updateCellValue(int? col, int? row, int value) {
    if ((col != null) && (row != null)) {
      if (!_board[row][col].isFixed) {
        _board[row][col] = Cell(
          row: row,
          col: col,
          value: value,
          isFixed: false,
        );

        saveCurrentGameState();
        notifyListeners();
      }
    }
  }

  bool get showConflicts => _showConflicts;
  void updateShowConflicts(bool showConflicts) {
    _showConflicts = showConflicts;
    notifyListeners();
  }

  void toggleShowConflicts() {
    updateShowConflicts(!showConflicts);
    saveCurrentGameState();
  }

  bool get animationInProgress => _animationInProgress;
  void updateAnimationInProgress(bool animationInProgress) {
    _animationInProgress = animationInProgress;
    notifyListeners();
  }

  void initAnimatedBackground() {
    AnimatedBoard staticBoard = [];
    final int boardSideLength = _blockSizeHorizontal * _blockSizeVertical;
    for (int row = 0; row < boardSideLength; row++) {
      List<bool> line = [];
      for (int col = 0; col < boardSideLength; col++) {
        line.add(false);
      }
      staticBoard.add(line);
    }
    _boardAnimated = staticBoard;
  }

  void setAnimatedBackground(List animatedCellsPattern) {
    final int boardSideLength = _blockSizeHorizontal * _blockSizeVertical;
    for (int row = 0; row < boardSideLength; row++) {
      for (int col = 0; col < boardSideLength; col++) {
        _boardAnimated[row][col] = animatedCellsPattern[row][col];
      }
    }
    notifyListeners();
  }

  void resetAnimatedBackground() {
    final int boardSideLength = _blockSizeHorizontal * _blockSizeVertical;
    for (int row = 0; row < boardSideLength; row++) {
      for (int col = 0; col < boardSideLength; col++) {
        _boardAnimated[row][col] = false;
      }
    }
  }
}
