import 'dart:math';

import 'package:sudoku/entities/cell.dart';
import 'package:sudoku/provider/data.dart';
import 'package:sudoku/utils/random_pick_grid.dart';
import 'package:sudoku/utils/tools.dart';

class BoardUtils {
  static printGrid(Board cells, Board solvedCells) {
    String stringValues = '0123456789ABCDEFG';
    printlog('');
    printlog('-------');
    for (int rowIndex = 0; rowIndex < cells.length; rowIndex++) {
      String row = '';
      String rowSolved = '';
      for (int colIndex = 0; colIndex < cells[rowIndex].length; colIndex++) {
        row += stringValues[cells[rowIndex][colIndex].value];
        rowSolved += stringValues[solvedCells[rowIndex][colIndex].value];
      }
      printlog('$row | $rowSolved');
    }
    printlog('-------');
    printlog('');
  }

  static Future<void> pickGrid(Data myProvider) async {
    String grid = '';
    RandomPickGrid randomPickGrid;

    randomPickGrid = RandomPickGrid();
    await randomPickGrid.init(myProvider.parameterLevel, myProvider.parameterSize);

    grid = randomPickGrid.grid;

    final int blockSizeHorizontal = myProvider.blockSizeHorizontal;
    final int blockSizeVertical = myProvider.blockSizeVertical;

    if (grid.length == pow(blockSizeHorizontal * blockSizeVertical, 2)) {
      printlog('Picked grid from template: $grid');
      bool isSymetric = (blockSizeHorizontal == blockSizeVertical);
      myProvider.updateCells(BoardUtils.createBoardFromTemplate(grid, isSymetric));
      myProvider.updateCellsSolved(BoardUtils.getSolvedGrid(myProvider));
      myProvider.selectCell(null, null);

      printGrid(myProvider.board, myProvider.boardSolved);
    }
  }

  static Board createEmptyBoard(final int boardSize) {
    final Board cells = [];
    for (int rowIndex = 0; rowIndex < boardSize; rowIndex++) {
      final List<Cell> row = [];
      for (int colIndex = 0; colIndex < boardSize; colIndex++) {
        row.add(Cell(
          row: rowIndex,
          col: colIndex,
          value: 0,
          isFixed: false,
        ));
      }
      cells.add(row);
    }

    return cells;
  }

  static Board createBoardFromSavedState(Data myProvider, String savedBoard) {
    final Board cells = [];
    final int boardSize = int.parse(pow((savedBoard.length / 2), 1 / 2).toStringAsFixed(0));

    const String stringValues = '0123456789ABCDEFG';

    int index = 0;
    for (int rowIndex = 0; rowIndex < boardSize; rowIndex++) {
      final List<Cell> row = [];
      for (int colIndex = 0; colIndex < boardSize; colIndex++) {
        final String stringValue = savedBoard[index++];
        final int value = stringValues.indexOf(stringValue);

        final String isFixedString = savedBoard[index++];
        final bool isFixed = (isFixedString != ' ');

        row.add(Cell(
          row: rowIndex,
          col: colIndex,
          value: value,
          isFixed: isFixed,
        ));
      }
      cells.add(row);
    }

    return cells;
  }

  static Board copyBoard(List cells) {
    final Board copiedGrid = [];
    for (int rowIndex = 0; rowIndex < cells.length; rowIndex++) {
      final List<Cell> row = [];
      for (int colIndex = 0; colIndex < cells[rowIndex].length; colIndex++) {
        row.add(Cell(
          row: rowIndex,
          col: colIndex,
          value: cells[rowIndex][colIndex].value,
          isFixed: false,
        ));
      }
      copiedGrid.add(row);
    }

    return copiedGrid;
  }

  static Board createBoardFromTemplate(String grid, bool isSymetric) {
    Board cells = [];
    final int boardSize = int.parse(pow(grid.length, 1 / 2).toStringAsFixed(0));

    const String stringValues = '0123456789ABCDEFG';

    int index = 0;
    for (int rowIndex = 0; rowIndex < boardSize; rowIndex++) {
      final List<Cell> row = [];
      for (int colIndex = 0; colIndex < boardSize; colIndex++) {
        final String stringValue = grid[index++];
        final int value = stringValues.indexOf(stringValue);
        row.add(Cell(
          row: rowIndex,
          col: colIndex,
          value: value,
          isFixed: (value != 0),
        ));
      }
      cells.add(row);
    }

    const List<String> allowedFlip = ['none', 'horizontal', 'vertical'];
    List<String> allowedRotate = ['none', 'left', 'right'];

    // Forbid rotation if blocks are not symetric
    if (!isSymetric) {
      allowedRotate = ['none'];
    }

    final Random rand = Random();
    final String flip = allowedFlip[rand.nextInt(allowedFlip.length)];
    final String rotate = allowedRotate[rand.nextInt(allowedRotate.length)];

    printlog('flip board: $flip');
    printlog('rotate board: $rotate');

    switch (flip) {
      case 'horizontal':
        {
          final Board transformedBoard = createEmptyBoard(boardSize);
          for (int rowIndex = 0; rowIndex < boardSize; rowIndex++) {
            for (int colIndex = 0; colIndex < boardSize; colIndex++) {
              transformedBoard[rowIndex][colIndex] = Cell(
                row: rowIndex,
                col: colIndex,
                value: cells[boardSize - rowIndex - 1][colIndex].value,
                isFixed: false,
              );
            }
          }
          cells = transformedBoard;
        }
        break;
      case 'vertical':
        {
          final Board transformedBoard = createEmptyBoard(boardSize);
          for (int rowIndex = 0; rowIndex < boardSize; rowIndex++) {
            for (int colIndex = 0; colIndex < boardSize; colIndex++) {
              transformedBoard[rowIndex][colIndex] = Cell(
                row: rowIndex,
                col: colIndex,
                value: cells[rowIndex][boardSize - colIndex - 1].value,
                isFixed: false,
              );
            }
          }
          cells = transformedBoard;
        }
        break;
    }

    switch (rotate) {
      case 'left':
        {
          final Board transformedBoard = createEmptyBoard(boardSize);
          for (int rowIndex = 0; rowIndex < boardSize; rowIndex++) {
            for (int colIndex = 0; colIndex < boardSize; colIndex++) {
              transformedBoard[rowIndex][colIndex] = Cell(
                row: rowIndex,
                col: colIndex,
                value: cells[colIndex][boardSize - rowIndex - 1].value,
                isFixed: false,
              );
            }
          }
          cells = transformedBoard;
        }
        break;
      case 'right':
        {
          final Board transformedBoard = createEmptyBoard(boardSize);
          for (int rowIndex = 0; rowIndex < boardSize; rowIndex++) {
            for (int colIndex = 0; colIndex < boardSize; colIndex++) {
              transformedBoard[rowIndex][colIndex] = Cell(
                row: rowIndex,
                col: colIndex,
                value: cells[boardSize - colIndex - 1][rowIndex].value,
                isFixed: false,
              );
            }
          }
          cells = transformedBoard;
        }
        break;
    }

    // Fix cells fixed states
    for (int rowIndex = 0; rowIndex < boardSize; rowIndex++) {
      for (int colIndex = 0; colIndex < boardSize; colIndex++) {
        cells[rowIndex][colIndex] = Cell(
          row: rowIndex,
          col: colIndex,
          value: cells[rowIndex][colIndex].value,
          isFixed: (cells[rowIndex][colIndex].value != 0) ? true : false,
        );
      }
    }

    return cells;
  }

  static bool checkBoardIsSolved(Data myProvider) {
    final Board cells = myProvider.board;
    final ConflictsCount conflicts = myProvider.boardConflicts;

    final int blockSizeHorizontal = myProvider.blockSizeHorizontal;
    final int blockSizeVertical = myProvider.blockSizeVertical;

    final int boardSize = blockSizeHorizontal * blockSizeVertical;

    // (re)compute conflicts
    BoardUtils.computeConflictsInBoard(myProvider);

    // check grid is fully completed and does not contain conflict
    for (int row = 0; row < boardSize; row++) {
      for (int col = 0; col < boardSize; col++) {
        if (cells[row][col].value == 0 || conflicts[row][col] != 0) {
          return false;
        }
      }
    }

    printlog('-> ok sudoku solved!');

    return true;
  }

  static bool isValueAllowed(
    Board cells,
    int blockSizeHorizontal,
    int blockSizeVertical,
    int? candidateCol,
    int? candidateRow,
    int candidateValue,
  ) {
    if ((candidateCol == null) || (candidateRow == null) || (candidateValue == 0)) {
      return true;
    }

    final int boardSize = blockSizeHorizontal * blockSizeVertical;

    // check lines does not contains a value twice
    for (int row = 0; row < boardSize; row++) {
      final List<int> values = [];
      for (int col = 0; col < boardSize; col++) {
        int value = cells[row][col].value;
        if (row == candidateRow && col == candidateCol) {
          value = candidateValue;
        }
        if (value != 0) {
          values.add(value);
        }
      }
      final List<int> distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        return false;
      }
    }

    // check columns does not contains a value twice
    for (int col = 0; col < boardSize; col++) {
      final List<int> values = [];
      for (int row = 0; row < boardSize; row++) {
        int value = cells[row][col].value;
        if (row == candidateRow && col == candidateCol) {
          value = candidateValue;
        }
        if (value != 0) {
          values.add(value);
        }
      }
      final List<int> distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        return false;
      }
    }

    // check blocks does not contains a value twice
    final int horizontalBlocksCount = blockSizeVertical;
    final int verticalBlocksCount = blockSizeHorizontal;
    for (int blockRow = 0; blockRow < verticalBlocksCount; blockRow++) {
      for (int blockCol = 0; blockCol < horizontalBlocksCount; blockCol++) {
        final List<int> values = [];

        for (int rowInBlock = 0; rowInBlock < blockSizeVertical; rowInBlock++) {
          for (int colInBlock = 0; colInBlock < blockSizeHorizontal; colInBlock++) {
            final int row = (blockRow * blockSizeVertical) + rowInBlock;
            final int col = (blockCol * blockSizeHorizontal) + colInBlock;
            int value = cells[row][col].value;
            if (row == candidateRow && col == candidateCol) {
              value = candidateValue;
            }
            if (value != 0) {
              values.add(value);
            }
          }
        }

        final List<int> distinctValues = values.toSet().toList();
        if (values.length != distinctValues.length) {
          return false;
        }
      }
    }

    return true;
  }

  static void computeConflictsInBoard(Data myProvider) {
    final Board cells = myProvider.board;
    final ConflictsCount conflicts = myProvider.boardConflicts;

    final int blockSizeHorizontal = myProvider.blockSizeHorizontal;
    final int blockSizeVertical = myProvider.blockSizeVertical;

    final int boardSize = blockSizeHorizontal * blockSizeVertical;

    // reset conflict states
    for (int row = 0; row < boardSize; row++) {
      for (int col = 0; col < boardSize; col++) {
        conflicts[row][col] = 0;
      }
    }

    // check lines does not contains a value twice
    for (int row = 0; row < boardSize; row++) {
      final List<int> values = [];
      for (int col = 0; col < boardSize; col++) {
        int value = cells[row][col].value;
        if (value != 0) {
          values.add(value);
        }
      }
      final List<int> distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        printlog('line $row contains duplicates');
        // Add line to cells in conflict
        for (int col = 0; col < boardSize; col++) {
          conflicts[row][col]++;
        }
      }
    }

    // check columns does not contains a value twice
    for (int col = 0; col < boardSize; col++) {
      final List<int> values = [];
      for (int row = 0; row < boardSize; row++) {
        int value = cells[row][col].value;
        if (value != 0) {
          values.add(value);
        }
      }
      final List<int> distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        printlog('column $col contains duplicates');
        // Add column to cells in conflict
        for (int row = 0; row < boardSize; row++) {
          conflicts[row][col]++;
        }
      }
    }

    // check blocks does not contains a value twice
    final int horizontalBlocksCount = blockSizeVertical;
    final int verticalBlocksCount = blockSizeHorizontal;
    for (int blockRow = 0; blockRow < verticalBlocksCount; blockRow++) {
      for (int blockCol = 0; blockCol < horizontalBlocksCount; blockCol++) {
        List<int> values = [];

        for (int rowInBlock = 0; rowInBlock < blockSizeVertical; rowInBlock++) {
          for (int colInBlock = 0; colInBlock < blockSizeHorizontal; colInBlock++) {
            int row = (blockRow * blockSizeVertical) + rowInBlock;
            int col = (blockCol * blockSizeHorizontal) + colInBlock;
            int value = cells[row][col].value;
            if (value != 0) {
              values.add(value);
            }
          }
        }

        List<int> distinctValues = values.toSet().toList();
        if (values.length != distinctValues.length) {
          printlog('block [$blockCol,$blockRow] contains duplicates');
          // Add blocks to cells in conflict
          for (int rowInBlock = 0; rowInBlock < blockSizeVertical; rowInBlock++) {
            for (int colInBlock = 0; colInBlock < blockSizeHorizontal; colInBlock++) {
              int row = (blockRow * blockSizeVertical) + rowInBlock;
              int col = (blockCol * blockSizeHorizontal) + colInBlock;
              conflicts[row][col]++;
            }
          }
        }
      }
    }

    myProvider.updateConflicts(conflicts);
  }

  static List<List<int>> getCellsWithWrongValue(
    final Board cells,
    final Board cellsSolved,
    final int blockSizeHorizontal,
    final int blockSizeVertical,
  ) {
    final List<List<int>> cellsWithWrongValue = [];
    final int boardSize = blockSizeHorizontal * blockSizeVertical;

    for (int row = 0; row < boardSize; row++) {
      for (int col = 0; col < boardSize; col++) {
        if (cells[row][col].value != 0 &&
            cells[row][col].value != cellsSolved[row][col].value) {
          cellsWithWrongValue.add([col, row]);
        }
      }
    }

    return cellsWithWrongValue;
  }

  static List<List<int>> getCellsWithUniqueAvailableValue(
    Board cells,
    final int blockSizeHorizontal,
    final int blockSizeVertical,
  ) {
    final List<List<int>> candidateCells = [];
    final int boardSize = blockSizeHorizontal * blockSizeVertical;

    for (int row = 0; row < boardSize; row++) {
      for (int col = 0; col < boardSize; col++) {
        if (cells[row][col].value == 0) {
          int allowedValuesCount = 0;
          int candidateValue = 0;
          for (int value = 1; value <= boardSize; value++) {
            if (BoardUtils.isValueAllowed(
                cells, blockSizeHorizontal, blockSizeVertical, col, row, value)) {
              candidateValue = value;
              allowedValuesCount++;
            }
          }
          if (allowedValuesCount == 1) {
            candidateCells.add([col, row, candidateValue]);
          }
        }
      }
    }

    return candidateCells;
  }

  static Board getSolvedGrid(Data myProvider) {
    final Board cells = copyBoard(myProvider.board);
    final int blockSizeHorizontal = myProvider.blockSizeHorizontal;
    int blockSizeVertical = myProvider.blockSizeVertical;

    do {
      List<List<int>> cellsWithUniqueAvailableValue =
          BoardUtils.getCellsWithUniqueAvailableValue(
        cells,
        blockSizeHorizontal,
        blockSizeVertical,
      );
      if (cellsWithUniqueAvailableValue.isEmpty) {
        break;
      }
      for (int i = 0; i < cellsWithUniqueAvailableValue.length; i++) {
        int col = cellsWithUniqueAvailableValue[i][0];
        int row = cellsWithUniqueAvailableValue[i][1];
        int value = cellsWithUniqueAvailableValue[i][2];
        cells[row][col] = Cell(
          row: row,
          col: col,
          value: value,
          isFixed: cells[row][col].isFixed,
        );
      }
    } while (true);

    return cells;
  }
}
